# -*- coding: utf-8 -*-

"""
Graph Tools.
"""

import numpy as np
import networkx as nx


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


# --------------------------------------------------------------------------
#                            GENERAL GRAPH TOOLS
# --------------------------------------------------------------------------

def graphs_equal(graph1, graph2):
    """Returns whether two NetworkX graphs (directed or undirected) are equal."""
    
    if (not graph1.order() == graph2.order()) or (not graph1.size() == graph2.size()):
        return False
    
    for u, v in graph1.edges():
        if not graph2.has_edge(u,v):
            return False
    
    return True


def performance(true_graph, graph):
    """Returns order, size, tp, tn, fp, fn, accuracy, precision and recall
    of a (directed or undirected) graph w.r.t. true graph."""
    
    order, size, tp, fp, fn = graph.order(), graph.size(), 0, 0, 0
    
    for u, v in graph.edges():
        if true_graph.has_edge(u,v):
            tp += 1
        else:
            fp += 1
    
    for u, v in true_graph.edges():
        if not graph.has_edge(u, v):
            fn += 1
    
    if isinstance(graph, nx.DiGraph):
        tn = (order * (order-1)) - (tp + fp + fn)
    else:
        tn = (order * (order-1) / 2) - (tp + fp + fn)
    
    accuracy = (tp + tn) / (tp + tn + fp + fn) if tp + tn + fp + fn > 0 else float('nan')
    precision = tp / (tp + fp) if tp + fp > 0 else float('nan')
    recall = tp / (tp + fn) if tp + fn > 0 else float('nan')
#    accuracy = (tp + tn) / (tp + tn + fp + fn) if tp + tn + fp + fn > 0 else 1.0
#    precision = tp / (tp + fp) if tp + fp > 0 else 1.0
#    recall = tp / (tp + fn) if tp + fn > 0 else 1.0
    
    return order, size, tp, tn, fp, fn, accuracy, precision, recall


def build_adjacency_matrix(G):
    """Return an adjacency matrix."""
    
    index = {node: i for i, node in enumerate(G.nodes())}   # maps node --> row/column index
    M = np.zeros( (len(index), len(index)), dtype=np.int8)
    for x, neighbors in G.adjacency():
        for y in neighbors:
            M[index[x], index[y]] = 1                       # add out-neighbor key_j to key_i
    return M


# --------------------------------------------------------------------------
#                             P4 EDITING
# --------------------------------------------------------------------------
    
def list_path(t, k, path, G, P4_list):
    for u in G[path[-1]]:
        G.nodes[u]['vis'] += 1
    for u in G[path[-1]]:
        if G.nodes[u]['vis'] == 1:
            if t < k:
                list_path(t+1, k, path + [u], G, P4_list)
            elif path[0] < u:
                P4_list.append(tuple(path) + (u,))
    for u in G[path[-1]]:
        G.nodes[u]['vis'] -= 1


def find_all_P4(G):
    P4_list = []
    for v in G:
            G.nodes[v]['vis'] = 0
    for v in G:
        G.nodes[v]['vis'] = 1
        list_path(2, 4, [v], G, P4_list)
        G.nodes[v]['vis'] = 0
    return P4_list


def is_good_quartet_OLD(path, BMG):
    """Determine whether an induced P4 is a good quartet in the BMG."""
    if (BMG.has_edge(path[0], path[2]) and 
        BMG.has_edge(path[3], path[1]) and
        BMG.nodes[path[0]]['color'] == BMG.nodes[path[3]]['color']):
        return True
    else:
        return False


def is_good_quartet(path, BMG):
    """Determine whether an induced P4 is a good quartet in the BMG."""
    induced = BMG.subgraph(path)
    degrees = [(induced.out_degree(v), induced.in_degree(v)) for v in induced]
    if (degrees[0] == (2,1) and 
        degrees[1] == (2,3) and
        degrees[2] == (2,3) and 
        degrees[3] == (2,1) and
        BMG.nodes[path[0]]['color'] == BMG.nodes[path[3]]['color']):
        return True
    else:
        return False
    

def remove_all_P4(RBMG, BMG, P4_list=None):
    """Find all good quartets and removes the middle edges."""
    GP4 = RBMG.copy()
    
    if P4_list is None:
        P4_list = find_all_P4(GP4)

    for path in P4_list:
        if is_good_quartet(path, BMG) and GP4.has_edge(path[1], path[2]):
            GP4.remove_edge(path[1], path[2])
            print(path[1], "-|-", path[2])
    
    return GP4

# --------------------------------------------------------------------------
#                            C6 DETECTION
# --------------------------------------------------------------------------

def validate_C6(G, C6_path):
    
    C6 = G.subgraph(C6_path)
    if C6.size() != 6:
        print(C6_path)
        print('Nodes', [n for n in C6.nodes()], C6.order())
        print('Edges', [edge for edge in C6.edges()], C6.size())
        print("Size not equal to 6!")
        return False
    for n in C6.nodes():
        if not C6.degree(n) == 2:
            print("Degree not equal to 2 for", n)
            return False
        for neighbor in G.neighbors(n):
            if G.nodes[n]['color'] == G.nodes[neighbor]['color']:
                print("Same color", n, neighbor)
                return False
    return True


def find_all_C6(G, P4_list=None):
    """Find induces C6 of the form <x1 y1 z1 x2 y2 z2>."""
    
    if P4_list is None:
        P4_list = find_all_P4(G)
        
    endpoints = {}
    for path in P4_list:
        a, b, c, d = path if path[0] < path[3] else path[::-1]
        a_col, b_col, c_col, d_col = (G.nodes[a]['color'], G.nodes[b]['color'],
                                      G.nodes[c]['color'], G.nodes[d]['color'])
        # avoid multiple recording for the 3 different colors:
        if a_col == d_col and a_col < b_col and a_col < c_col:
            if (a, d) not in endpoints:
                endpoints[(a, d)] = {}
            if (b_col, c_col) not in endpoints[(a, d)]:
                endpoints[(a, d)][(b_col, c_col)] = []
            endpoints[(a, d)][(b_col, c_col)].append(path)
    
    C6_list = []
    for endpoint, colors in endpoints.items():
        x1, x2 = endpoint
        for color_pair, paths in colors.items():
            y_color, z_color = color_pair
            # avoid multiple recording for the 2 directions:
            if y_color < z_color and (z_color, y_color) in colors:
                for path1 in paths:
                    for path2 in colors[(z_color, y_color)]:
                        if ((not G.has_edge(path1[1], path2[1])) and
                            (not G.has_edge(path1[2], path2[2]))):
                            C6_list.append(path1 + (path2[2], path2[1]))
                            if not validate_C6(G, C6_list[-1]):
                                raise Exception("Problem in C6 detection")
                            print(C6_list[-1])
    return C6_list


def graph_type(G):
    
    P4_list = find_all_P4(G)
    C6_list = find_all_C6(G, P4_list=P4_list)
    
    if C6_list:
        return "C", P4_list, C6_list
    elif P4_list:
        return "B", P4_list, C6_list
    else:
        return "A", P4_list, C6_list



# -*- coding: utf-8 -*-

import os, itertools, random

import numpy as np
import networkx as nx
from Bio import SeqIO


from asymmetree.POParser import parse_best_match_candidates
from asymmetree.POOutgroupFinder import OutgroupFinder
from asymmetree.PODistanceCalculator import distance_2seqs

class POBestMatches:
    
    def __init__(self, fasta_files, candidate_dir,
                 tree_file,
                 epsilon=0.5,
                 voting_mode="majority"):
        
        self.BMG = nx.DiGraph()
        self.fasta_files = fasta_files
        self.candidate_dir = candidate_dir
        
        self.tree_file = tree_file
        
        self.epsilon = epsilon
        
        if voting_mode in ("majority", "weighted"):
            self.voting_mode = voting_mode
        else:
            raise ValueError(f"Invalid voting mode for quartets '{voting_mode}'")
        
    
    def __call__(self):
        
        self._check_file_existence()
        self._build_fasta_index()
        
        self.distance_cache = {}
        
        self.outgroup_finder = OutgroupFinder(self.tree_file)
        
        for spec in self.species.keys():
            print("----- Species '{}' -----".format(spec))
            self._best_matches_for_species(spec)
            
        return self.BMG
    
    
    def _check_file_existence(self):
        
        self.species = {}
        self.candidate_files = {}
        
        # check if all fasta files exist
        for filename in self.fasta_files:
            
            if os.path.isfile(filename):
                self.species[os.path.basename(filename)] = filename
            else:
                raise FileNotFoundError(f"Could not find file '{filename}'.")
        
        # check if the best match candidate files exist
        for spec in self.species.keys():
            
            filename = os.path.join(self.candidate_dir, spec + ".bm_candidates")
            if os.path.isfile(filename):
                self.candidate_files[spec] = filename
            else:
                raise FileNotFoundError(f"Could not find file '{filename}'.")
                
        # check if the tree file exists
        if self.tree_file is not None and not os.path.isfile(self.tree_file):
            raise FileNotFoundError(f"Could not find tree file '{self.tree_file}'.")
    
    
    def _build_fasta_index(self):
        
        self.fasta_index = {}
        
        for spec, filename in self.species.items():
            self.fasta_index[spec] = SeqIO.index(filename, "fasta")
            
            for gene_specifier in self.fasta_index[spec].keys():
                self.BMG.add_node("{}_{}".format(spec, gene_specifier),
                                  gene_specifier=gene_specifier,
                                  color=spec)

    
    def _get_seq_from_label(self, label, spec):
        
#        # original label in fasta file (and index) without 'SPECIES_'
#        label = label[len(spec)+1:]
        
        return self.fasta_index[spec][label].seq
    
    
    def _best_matches_for_species(self, spec_x):
        
        candidates = parse_best_match_candidates(self.candidate_files[spec_x])
        
        i = 1
        for x in candidates.keys():
            
            print("----- Processing gene {}/{} -----".format(i, len(candidates)))
            i += 1
            
            x_label = "{}_{}".format(spec_x, x)
            
            for spec_Y, Y in candidates[x].items():
                
                # trivial case: |Y| = 1
                if len(Y) == 1:
                    best_matches = [ Y[0][1] ]
                else:
                    outgroups = self.outgroup_finder(spec_x, spec_Y, candidates[x],
                                                     limit=10)
                    
                    if outgroups:
                        best_matches = self._quartet_method(x, spec_x, Y, spec_Y,
                                                            outgroups)
                    else:
                        best_matches = self._ext_best_hits(x, spec_x, Y, spec_Y)
                
                for bm in best_matches:
                    y_label = "{}_{}".format(spec_Y, bm)
                    self.BMG.add_edge(x_label, y_label)
                    
                    
    def _get_distance(self, a, spec_a, b, spec_b):
        
        # pairwise distances are symmetric --> compute just once
        if (a, spec_a) < (b, spec_b):
            identifier = (a, spec_a, b, spec_b)
        else:
            identifier = (b, spec_b, a, spec_a)
        
        if identifier in self.distance_cache:
            return self.distance_cache[identifier]
        else:
            distance = distance_2seqs(self.fasta_index[spec_a][a].seq,
                                      self.fasta_index[spec_b][b].seq)
#            distance = random.random()
            self.distance_cache[identifier] = distance
            return distance
        
    
    def _ext_best_hits(self, x, spec_x, Y, spec_Y):
        
        best_matches = []
        
        distances, min_distance = [], float("inf")
        for _, y, _, _, _ in Y:
            distance_to_y = self._get_distance(x, spec_x, y, spec_Y)
            distances.append(distance_to_y)
            if distance_to_y < min_distance:
                min_distance = distance_to_y
                
        for i in range(len(Y)):
            if distances[i] < (1 + self.epsilon) * min_distance:
                best_matches.append(Y[i][1])
        
        return best_matches
    
    
    def _quartet_method(self, x, spec_x, Y, spec_Y, Z):
        
        H = nx.DiGraph()
#        f = 1 - self.insecurity_factor
        
        for y1, y2 in itertools.combinations(Y, 2):
            
            y1, y2 = y1[1], y2[1]
            
            votes = [0, 0, 0, 0]
            for spec_z, z, _, _, _ in Z:
                q, weight = self._supported_quartet(x,      spec_x,
                                                    y1, y2, spec_Y,
                                                    z,      spec_z)
                if self.voting_mode == "weighted sum":
                    votes[q] += weight
                elif self.voting_mode == "majority":
                    votes[q] += 1
            
            sorted_votes = sorted((vote, i) for i, vote in enumerate(votes))
            
            if sorted_votes[2][0] > 0.9 * sorted_votes[3][0]:   # not sure if the two 
                quartet = 3                                     # top votes are too close
                                                                # default is star topology
            else:
                quartet = sorted_votes[3][1]
            
            if quartet == 0:                                    # 0: xy1 | y2z
                H.add_edge(y2, y1)
            elif quartet == 1:                                  # 1: xy2 | y1z
                H.add_edge(y1, y2)
            else:                                               # 2: xz | y1y2   or   3: star
                H.add_edge(y1, y2)
                H.add_edge(y2, y1)
        
        # find the (unique) strongly connected component without out-edges
        sccs = {i: scc for i, scc in enumerate(nx.strongly_connected_components(H))}
        
        for i, scc in sccs.items():
            for y in scc:
                H.nodes[y]["scc"] = i
        while sccs:
            for i, scc in sccs.items():
                out_edges = False
                for y1 in scc:
                    for y2 in H.successors(y1):
                        if H.nodes[y1]["scc"] != H.nodes[y2]["scc"]:
                            out_edges = True
                            break
                    if out_edges:
                        break
                if not out_edges:
                    return list(scc)
    
    
    def _supported_quartet(self, x,      spec_x,
                                 y1, y2, spec_y,
                                 z,      spec_z):
        
        xy1_y2z = (self._get_distance(x,  spec_x, y1, spec_y) +
                   self._get_distance(y2, spec_y, z,  spec_z))
        xy2_y1z = (self._get_distance(x,  spec_x, y2, spec_y) +
                   self._get_distance(y1, spec_y, z,  spec_z))
        xz_y1y2 = (self._get_distance(x,  spec_x, z,  spec_z) +
                   self._get_distance(y1, spec_y, y2, spec_y))
        
        if xy1_y2z < xy2_y1z and xy1_y2z < xz_y1y2:     # 0: xy1 | y2z
            quartet = 0
        elif xy2_y1z < xy1_y2z and xy2_y1z < xz_y1y2:   # 1: xy2 | y1z
            quartet = 1
        elif xz_y1y2 < xy1_y2z and xz_y1y2 < xy2_y1z:   # 2: xz | y1y2
            quartet = 2
        else:                                           # 3: star
            quartet = 3
            
        d0, d1, d2 = sorted([xy1_y2z, xy2_y1z, xz_y1y2])
        weight = (1 - d0/d2) * np.exp(d1 - d2) if d2 > 0 else 0.0
        
        return quartet, weight
    
    
def print_BMG(BMG, filename):
    
    with open(filename, "w") as f:
        
        for u, v in BMG.edges:
            
            f.write("{}\t{}\t{}\t{}\n".format(BMG.nodes[u]["gene_specifier"],
                                              BMG.nodes[v]["gene_specifier"],
                                              BMG.nodes[u]["color"],
                                              BMG.nodes[v]["color"],))


# -*- coding: utf-8 -*-

"""
Doubly-linked list.

Implementation of a doubly-linked list in Python 3. The list enables access
to single list elements in order to modify/delete values in constant time.

Classes in this module:
    - DLListElement     (doubly-linked list element)
    - DLList            (doubly-linked list)
"""

import collections


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2020, David Schaller"


class DLListElement:
    """Doubly-linked list element."""
    
    __slots__ = ['value', 'prev_el', 'next_el']
    
    def __init__(self, value, prev_el=None, next_el=None):
        
        self.value = value
        self.prev_el = prev_el
        self.next_el = next_el


class DLList:
    """Doubly-linked list."""
    
    __slots__ = ['_first', '_last', 'count', 'current']
    
    def __init__(self, *args):
        
        self._first = None
        self._last = None
        self.count = 0
        for arg in args:
            if isinstance(arg, collections.Iterable):
                for item in arg:
                    self.append(item)
            else:
                self.append(item)
        
    
    def __len__(self):
        
        return self.count
    
    
    def __nonzero__(self):
        
        return True if self.count > 0 else False
    
    
    def __iter__(self):
        
        return DLListIterator(self)


    def __next__(self):
        
        pass
    
    
    def __getitem__(self, index):
        
        if not isinstance(index, int):
            raise TypeError("Index must be of type 'int'!")
        if index >= 0:
            if index >= self.count:
                raise IndexError("Index " + str(index) + " is out of bounds!")
            element = self._first
            for i in range(index):
                element = element.next_el
            return element.value
        else:
            if index < (-self.count):
                raise IndexError("Index " + str(index) + " is out of bounds!")
            element = self._last
            for i in range(-index - 1):
                element = element.prev_el
            return element.value
        raise IndexError("Index " + str(index) + " is out of bounds!")
    
    
    def first(self):
        
        return self._first.value
    
    
    def last(self):
        
        return self._last.value
    
    
    def first_element(self):
        
        return self._first
    
    
    def append(self, value):
        
        new_end = DLListElement(value, prev_el=self._last)
        if self._last:
            self._last.next_el = new_end
        self._last = new_end
        if not self._first:
            self._first = new_end
        self.count += 1
        return new_end
    
    
    def extend(self, iterable):
        
        for value in iterable:
            self.append(value)
    
    
    def append_left(self, value):
        
        new_start = DLListElement(value, next_el=self._first)
        if self._first:
            self._first.prev_el = new_start
        self._first = new_start
        if not self._last:
            self._last = new_start
        self.count += 1
        return new_start
    
    
    def remove_element(self, element):
        """Remove an item by reference to the 'DLListElement' instance in O(1)."""
        
        if element.prev_el:
            element.prev_el.next_el = element.next_el
        if element.next_el:
            element.next_el.prev_el = element.prev_el
        if self._first is element:
            self._first = element.next_el
        if self._last is element:
            self._last = element.prev_el
        element.prev_el, element.next_el = None, None
        self.count -= 1
    
    
    def remove(self, value):
        """Remove an item by value in O(n)."""
        
        element = self._first
        while element:
            if element.value == value:
                self.remove_element(element)
                return
            element = element.next_el
        raise KeyError("Value", value, "is not in the doubly-linked list!")
    
    
    def popright(self):
        """Removes the last element of the list and returns its value."""
        
        if self._last:
            value = self._last.value
            self.remove_element(self._last)
            return value
        else:
            return None
    
    
    def popleft(self):
        """Removes the first element of the list and returns its value."""
        
        if self._first:
            value = self._first.value
            self.remove_element(self._first)
            return value
        else:
            return None
    
    
    def clear(self):
        
        self._first = None
        self._last = None
        self.count = 0
        

class DLListIterator:
    """Iterator class for doubly-linked list."""
    
    def __init__(self, dllist):
        
        self.dllist = dllist
        self._current = dllist._first
        
    
    def __next__(self):
        
        if self._current:
            x = self._current
            self._current = self._current.next_el
            return x.value
        else:
            raise StopIteration
        

if __name__ == "__main__":
    dllist = DLList([7, 8, 9])
    dllist.append(1)
    dllist.append(2)
    dllist.append(3)
    dllist.append(4)
    dllist.append(5)
    dllist.remove(3)
    dllist.append_left(30)
    for x in dllist:
        print(x, end=" ")
    print("\n", dllist[0], dllist[1], dllist[2], dllist[3], dllist[-4])


# -*- coding: utf-8 -*-

"""
Construction of Least Resolved Trees (LRT).

Computes an LRT from a (not necessarily valid) cBMG or
from an observable gene tree.
"""

import itertools

import networkx as nx

from asymmetree.best_matches import TrueBMG
from asymmetree.tools.PhyloTree import PhyloTree, PhyloTreeNode


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2020, David Schaller"


# --------------------------------------------------------------------------
#                      LRT FROM OBSERVABLE GENE TREE
#
#        WARNING: seems to work, but not (yet) mathmatically proven
# --------------------------------------------------------------------------


def _rho_colors(T, subtree_colors):
    """Computes the colors s for which each v is the root of some thinness class."""
    
    colors = subtree_colors[T.root]                   # full color set
    
    rho_colors = {v: set() for v in T.preorder()}     # colors s for which v is the root of some thinness class
    
    for u in T.root.leaves:
        remaining = colors - {u.color}                  # colors to which no best match has yet been found
        current = u.parent                              # start with direct parent of each node
        while remaining and current:
            colors_here = set()
            for v in current.leaves:
                if v.color in remaining:                # best match found
                    colors_here.add(v.color)
            
            rho_colors[current] |= colors_here          # update thinness class root colors
            remaining -= colors_here                    # update remaining colors
            current = current.parent
    
    return rho_colors


def _redundant_edges(T, subtree_colors, rho_colors):
    
    redundant_edges = []
    
    for u, v in T.inner_edges():
        
        aux_set = set()                                 # colors s in sigma( L(T(u) \ T(v)) )
        for v2 in u.children:
            if v2 is not v:
                aux_set.update(subtree_colors[v2])
        
        if not rho_colors[v].intersection(aux_set):
            redundant_edges.append((u, v))
    
    return redundant_edges


def LRT_from_observable_tree(T):
    """Computes the Least Resolved Tree from a tree.
    
    The unique Least Resolved Tree from a leaf-colored (observable)
    gene tree is computed by contraction of all redundant edges.
    
    WARNING: seems to work, but not (yet) mathmatically proven!
    """
    
    LRT = T.copy()
    
    # remove planted root
    if len(LRT.root.children) == 1:
        LRT.delete_and_reconnect(LRT.root.children[0])
    
    LRT.supply_leaves()                                 # assign list of leaves to each node
    
    subtree_colors = {}
    for v in LRT.preorder():
        subtree_colors[v] = {leaf.color for leaf in v.leaves}
        
    rho_colors = _rho_colors(LRT, subtree_colors)
    redundant_edges = _redundant_edges(LRT, subtree_colors, rho_colors)
    LRT.contract(redundant_edges)
    LRT = LRT.topology_only()
    
    return LRT


# --------------------------------------------------------------------------
#                               LRT FROM BMG
# --------------------------------------------------------------------------

class LRTConstructor:
    
    def __init__(self, G, mincut=True):
        
        self.G = G
        self.R = []
        self.L = set()
        self.color_dict = {}
        self.mincut = mincut
        self.cut_value = 0
        self.cut_list = []
        
    
    def initialize(self):
        """Initialize color dictionary and leaf set."""
        
        for v in self.G.nodes:
            color = self.G.nodes[v]["color"]
            if color not in self.color_dict:
                self.color_dict[color] = [v]
            else:
                self.color_dict[color].append(v)
            self.L.add(v)
    
    
    def informative_triples(self):
        """Compute the informative triples."""
        
        for col1, col2 in itertools.permutations(self.color_dict.keys(), 2):
            list1, list2 = self.color_dict[col1], self.color_dict[col2]
            for a in list1:
                for b, c in itertools.permutations(list2, 2):
                    if self.G.has_edge(a,b) and (not self.G.has_edge(a,c)):   # X1 / X2 / X3 / X4
                        self.R.append( (a,b,c) )
    
    
    def build_tree(self):
        """Build the least resolved tree from informative triples."""
        
        self.initialize()
        self.informative_triples()
        
        root = self.aho(self.L, self.R)
        if not root:
            print("No such tree exists!")
            return None
        else:
            return PhyloTree(root)
    
    
    def aho(self, L, R):
        """Recursive Aho-algorithm."""
        
        if len(L) == 1:                                 # trivial case: only one leaf left in L
            leaf = L.pop()
            return PhyloTreeNode(leaf, label=self.G.nodes[leaf]["label"],
                            color=self.G.nodes[leaf]["color"])
            
        help_graph_A = HelpGraph(L, R, self)            # construct the help graph A(R)
                                                        # determine connected components A1, ..., Ak
        conn_comps = help_graph_A.connected_comp(mincut=self.mincut)
        
        if len(conn_comps) <= 1:                        # return False if less than 2 connected components
            print("Connected component:\n", conn_comps)
            return False
        child_nodes = []
        for cc in conn_comps:
            Li = set(cc)                                # list/dictionary --> set
            Ri = []
            for t in R:                                 # determine which triples are in the subtree
                if Li.issuperset(t):
                    Ri.append(t)
            Ti = self.aho(Li, Ri)                       # recursive call
            if not Ti:
                return False                            # raise False to previous call of aho()
            else:
                child_nodes.append(Ti)
                
        subtree_root = PhyloTreeNode(0, label="")       # place new inner node
        for Ti in child_nodes:
            subtree_root.add_child(Ti)                  # add roots of the subtrees to the new node
   
        return subtree_root                             # return the new node
    
    
    @staticmethod
    def correct_BMG(BMG_original):
        """Build the LRT (using Mincut in Aho algorithm and return its BMG and RBMG."""
        
        subtrees = []
        for sg in (BMG_original.subgraph(c) for c in nx.weakly_connected_components(BMG_original)):
            lrt_constructor = LRTConstructor(sg, mincut=True)
            tree = lrt_constructor.build_tree()
            if tree:
                subtrees.append(tree)
                
        if len(subtrees) == 0:
            return None
        elif len(subtrees) == 1:
            tree = subtrees[0]
        else:
            tree = PhyloTree(PhyloTreeNode(0))
            for subtree in subtrees:
                tree.root.add_child(subtree.root)
        
        return TrueBMG.best_match_graphs(tree)


class HelpGraph:
    """Help graph for Aho-algorithm.
    
    Takes a list of leaves L and a list of tuples R of the form (x,y|z) and
    builds a graph with V = L and E = { (x,y) | (x,y|z) in R }.
    Provides optional Minimal Edge Cut for function connected_comp().
    """
    
    def __init__(self, L, R, aho):
        """Constructs the help graph as NetworkX graph.
        
        Edges (a,b) are weighted by the number of occurrences in triples
        of the form ab|x or ba|x.
        """
        
        self.G = nx.Graph()
        self.G.add_nodes_from(L)
        self.aho = aho

        for t1, t2, t3 in R:
            if not self.G.has_edge(t1, t2):
                self.G.add_edge(t1, t2, weight=1)
#            if self.G.has_edge(t1, t2):
#                self.G[t1][t2]["weight"] += 1
#            else:
#                self.G.add_edge(t1, t2, weight=1)
    
    
    def connected_comp(self, mincut=False):
        """Determines the connected components of the graph.
        
        Keyword arguments:
            mincut -- if True and graph has only one component, a 
                      minimal edge cut is applied first, function then
                      always return > 1 components (default=False).
        """
        
        if (not mincut) or nx.number_connected_components(self.G) > 1:
            return list(nx.connected_components(self.G))
        else:
            cut = nx.stoer_wagner(self.G)           # Stoer–Wagner algorithm
            self.aho.cut_value += cut[0]
            if len(cut[1][0]) < len(cut[1][1]):
                smaller_comp = cut[1][0]
            else:
                smaller_comp = cut[1][1]
            for edge in self.G.edges:
                if ((edge[0] in smaller_comp and edge[1] not in smaller_comp) or
                    (edge[1] in smaller_comp and edge[0] not in smaller_comp)):
                    self.aho.cut_list.append(edge)
            return cut[1]
        
        
if __name__ == "__main__":
    
    colors = [i+1 for i in range(5)]
    T = PhyloTree.random_colored_tree(30, colors)
    print("--- T ---\n", T.to_newick())
    
    BMG, _ = TrueBMG.best_match_graphs(T)
    
    lrt_constr = LRTConstructor(BMG, mincut=False)
    LRT1 = lrt_constr.build_tree()
    
    LRT2 = LRT_from_observable_tree(T)
    
    print("--- LRT1 ---\n", LRT1.to_newick())
    print("--- LRT2 ---\n", LRT2.to_newick())
    
    print(f"LRTs equal: {LRT1.compare_topology(LRT2)}")


# -*- coding: utf-8 -*-

import itertools

import asymmetree.tools.DoublyLinkedList as dll


class TreeNode:
    """Class 'TreeNode'.
    
    Components for class 'Tree'. Contains a list of children as well as a
    reference to the parent node.
    """
    
    __slots__ = ['ID', 'label', 'parent', 'parent_dll_element',
                 'children', 'leaves']
    
    
    def __init__(self, ID, label=None):
        
        self.ID = ID
        self.label = label
        self.parent = None
        self.children = dll.DLList()
        self.parent_dll_element = None      # reference to doubly-linked list element
                                            # in the parents' children
    
    
    def __repr__(self):
        return "tn" + str(self.ID)
    
                                            
    def add_child(self, child_node):
        
        # do nothing if child_node is already a child of self
        
        if child_node.parent is None:
            child_node.parent = self
            child_node.parent_dll_element = self.children.append(child_node)
        
        elif child_node.parent is not self:
            child_node.parent.remove_child(child_node)
            child_node.parent = self
            child_node.parent_dll_element = self.children.append(child_node)  
    
    
    def remove_child(self, child_node):
        
        if child_node.parent is self:
            self.children.remove_element(child_node.parent_dll_element)
            child_node.parent = None
            child_node.parent_dll_element = None
        else:
            raise ValueError("Not a child of this node!")
            
            
    def detach(self):
        
        if self.parent is not None:
            self.parent.remove_child(self)
        else:
            self.parent = None
            self.parent_dll_element = None
        

class Tree:
    
    def __init__(self, root):
        self.root = root
    
    
    def preorder(self, node=None):
        """Generator for preorder traversal of the tree."""
        
        if node is None:
            yield from self.preorder(node=self.root)
        else:
            yield node
            for child in node.children:
                yield from self.preorder(node=child)
    
    
    def postorder(self, node=None):
        """Generator for postorder traversal of the tree."""
        
        if node is None:
            yield from self.postorder(node=self.root)
        else:
            for child in node.children:
                yield from self.postorder(node=child)
            yield node
            
    
    def inner_vertices(self, node=None):
        """Generator for inner vertices in preorder."""
        
        if node is None:
            yield from self.inner_vertices(node=self.root)
        else:
            if node.children:
                yield node
                for child in node.children:
                    yield from self.inner_vertices(node=child)
            
    
    def edges(self, node=None):
        """Generator for all edges of the tree."""
        
        if node is None:
            yield from self.edges(node=self.root)
        else:
            for child in node.children:
                yield (node, child)
                yield from self.edges(node=child)
                
    
    def inner_edges(self, node=None):
        """Generator for all inner edges of the tree."""
        
        if node is None:
            yield from self.inner_edges(node=self.root)
        else:
            for child in node.children:
                if child.children:
                    yield (node, child)
                    yield from self.inner_edges(node=child)
                
    
    def euler_generator(self, node=None, id_only=False):
        """Generator for an Euler tour of the tree."""
        
        if node is None:
            yield from self.euler_generator(node=self.root, id_only=id_only)
        else:
            if id_only: yield node.ID
            else: yield node
            
            for child in node.children:
                yield from self.euler_generator(node=child, id_only=id_only)
                
                if id_only: yield node.ID
                else: yield node
            
    
    def supply_leaves(self, node=None):
        """Add the leaves to all nodes that are in the subtree of a specific node."""
        
        if node is None:
            self.supply_leaves(self.root)
            return self.root.leaves
        else:
            node.leaves = []
            if not node.children:
                node.leaves.append(node)
            else:
                for child in node.children:
                    node.leaves.extend(self.supply_leaves(child))
            return node.leaves
        
        
    def get_triples(self):
        """Retrieve a list of all triples of the tree.
        
        Warning: Algorithm works recursively and can produce extremely 
        large lists.
        """
                
        self.supply_leaves()
        triples = []
        
        for u in self.preorder():
            for v1, v2 in itertools.permutations(u.children, 2):
                if len(v2.leaves) > 1:
                    for t3 in v1.leaves:
                        for t1, t2 in itertools.combinations(v2.leaves, 2):
                            triples.append( (t1, t2, t3) )
        
        return triples
    
    
    def delete_and_reconnect(self, node):
        """Delete a node from the tree and reconnect its parent and children."""
        
        parent = node.parent
        if not parent:
            print(f"Cannot delete and reconnect root '{node}'!")
            return False
        else:
            parent.remove_child(node)
            
            # copy list of children to edit edges
            children = [child for child in node.children]
            for child in children:
                parent.add_child(child)
                    
            node.children.clear()
        
        return parent
        
    
    def to_newick(self, node=None):
        """Recursive Tree --> Newick (str) function."""
        
        if node is None:
            return self.to_newick(self.root) + ";"
        elif not node.children:
            return str(node)
        else:
            s = ''
            for child in node.children:
                s += self.to_newick(node=child) + ","
            return f"({s[:-1]}){node}"
      
    
    def get_hierarchy(self):
        """Return the hierarchy set on the leaf labels defined by the tree."""
        
        self.supply_leaves()
        
        hierarchy = set()
        
        for v in self.preorder():
            
            A = [leaf.label for leaf in v.leaves]
            A.sort()
            A = tuple(A)
            hierarchy.add(A)
            
        return hierarchy
    
    
    def compare_topology(self, other):
        """Compare the tree topology based on the hierarchies.
        
        Only works for binary trees."""
        
        hierarchy1 = sorted(self.get_hierarchy())
        hierarchy2 = sorted(other.get_hierarchy())
        
        if len(hierarchy1) != len(hierarchy2):
            print(f"Unequal sizes of the hierarchy sets: {len(hierarchy1)} and {len(hierarchy2)}")
            return False
        
        for i in range(len(hierarchy1)):
            
#            if len(hierarchy1[i]) > 1:
#                print(hierarchy1[i])
#                print(hierarchy2[i])
            
            if hierarchy1[i] != hierarchy2[i]:
                print(f"Hierarchies not equal:\n{hierarchy1[i]}\n{hierarchy2[i]}")
                return False
        
        return True


# -*- coding: utf-8 -*-

"""
Tree Imbalancer.

Introduce evolution rate asymmetries and autocorrelation.

Methods in this module:
    - imbalance_tree
"""

import numpy as np

from asymmetree.tools.PhyloTree import PhyloTree


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


# --------------------------------------------------------------------------
#                      IMBALANCING OF THE GENE TREE
#
# --------------------------------------------------------------------------


def imbalance_tree(T, S, baseline_rate=1,
                   lognormal_v=0,
                   gamma_param=(0.5, 1.0, 2.2),
                   weights=(1/3, 1/3, 1/3), copy_tree=False):
    """Imbalances an (ultrametric) TRUE gene tree.
    
    Keyword arguments:
    baseline_rate   -- mean of substitution rate for conserved genes
    lognormal_v     -- variance factor for lognormal distribution
    gamma_param     -- param. for gamma distribution (a, loc, scale)
    weights         -- weights for choice between conservation,
                       subfunctionalization and neofunctionalization
    copy_tree       -- copy the tree before imbalancing
    """
    
    if copy_tree:
        T = T.copy()
    weights = np.asarray(weights) / sum(weights)
        
    _divergent_rates(T, S, gamma_param, weights)
    
    _autocorrelation(T, S, baseline_rate, lognormal_v)
    
    return T


# --------------------------------------------------------------------------
#                       EVOLUTION RATE ASYMMETRY
#
# --------------------------------------------------------------------------

def _adjust_distances(T, rates):
    for edge, rate_list in rates.items():
        time_points = np.asarray([tstamp for tstamp, _ in rate_list] + [edge[1].tstamp])
        rate_values = np.asarray([rate for _, rate in rate_list])
        edge[1].dist = np.dot(-np.diff(time_points),rate_values)


def _divergent_rates(T, S, gamma_param, weights):
    """
    Keyword arguments:
    gamma_param     -- param. for gamma distribution (a, loc, scale)
    weights         -- weights for choice between conservation,
                       subfunctionalization and neofunctionalization
    """
    
    def get_rate(marked_as):
        if marked_as == "conserved":
            # --- conserved genes are set to rate 1 ---
            return 1.0
        elif marked_as == "divergent":
            # --- fitted gamma distribution ---
            return (np.random.gamma(gamma_param[0], scale=gamma_param[2]) +
                    gamma_param[1])
    
    def duplication_type(marked_as):
        if marked_as == "divergent":
            return "divergent", "divergent"
        else:
            r = np.random.choice(3, p=weights)
            if r == 0:                                  # conservation
                return "conserved", "conserved"
            elif r == 1:                                # subfunctionalization
                return "divergent", "divergent"
            else:                                       # neofunctionalization
                if np.random.uniform() < 0.5:
                    return "divergent", "conserved"
                else:
                    return "conserved", "divergent"
    
    T_nodes = T.sorted_nodes()
    rates = {edge: [] for edge in T.edges()}        # edge --> list of (tstamp, rate) tuples
    
    S_parents = {v.ID: v.parent.ID for v in S.preorder() if v.parent}
    gene_counter = {(e[0].ID, e[1].ID): [] for e in S.edges()}
    marked = {v: "conserved" for v in T_nodes}      # marked as conserved or divergent
    
    for u in T_nodes:
        
        # ----------------- SPECIATION -----------------
        if u.label in ("S", ""):
            for v in u.children:
                marked[v] = marked[u]
                S_u = u.color
                S_v = v.color if not isinstance(v.color, tuple) else v.color[1]
                gene_counter[(S_u, S_v)].append(v)
                rates[(u,v)].append((u.tstamp, get_rate(marked[v])))
#                rates[(u,v)].append((u.tstamp, rates[(u.parent,u)][-1][1] if u.parent else 1.0))
            
        # ---------------- DUPLICATION -----------------
        elif u.label == "D":
            marked[u.children[0]], marked[u.children[1]] = duplication_type(marked[u])
            gene_counter[u.color].remove(u)
            for v in u.children:
                gene_counter[u.color].append(v)
                rates[(u,v)].append((u.tstamp, get_rate(marked[v])))
#                if marked[u] == "conserved":
#                    rates[(u,v)].append((u.tstamp, get_rate(marked[v])))
#                else:
#                    rates[(u,v)].append((u.tstamp, rates[(u.parent,u)][-1][1]))
        
        # ------------------- LOSS ---------------------
        elif u.label == "*":
            gene_counter[u.color].remove(u)
            if len(gene_counter[u.color]) == 1:
                v = gene_counter[u.color][0]
                if marked[v] == "divergent":
                    marked[v] = "conserved"
                    rates[(v.parent,v)].append((u.tstamp, get_rate(marked[v])))
        
        # ---------- HORIZONTAL GENE TRANSFER ----------
        elif u.label == "H":
            v1, v2 = u.children
            if v1.transferred:
                v1, v2 = v2, v1         # now v2 is the transferred copy
                
            # untransferred copy
            marked[v1] = marked[u]
            gene_counter[u.color].remove(u)
            gene_counter[u.color].append(v1)
            if u.parent:
                rates[(u,v1)].append((u.tstamp, rates[(u.parent,u)][-1][1]))
            else:
                rates[(u,v1)].append((u.tstamp, get_rate(marked[v1])))
            
            # transferred copy
            marked[v2] = "divergent"
            if isinstance(v2.color, tuple):
                gene_counter[v2.color].append(v2)
            else:
                gene_counter[(S_parents[v2.color], v2.color)].append(v2)
            rates[(u,v2)].append((u.tstamp, get_rate(marked[v2])))
            
    _adjust_distances(T, rates)
    return T

# --------------------------------------------------------------------------
#                         AUTOCORRELATION
#
# --------------------------------------------------------------------------

def _autocorrelation(T, S, baseline_rate, lognormal_v):
    """
    Keyword arguments:
    baseline_rate   -- mean of substitution rate for conserved genes
    lognormal_v     -- variance factor for lognormal distribution
    """
    rates = {}                                                          # maps node v --> rate
    S_parent = {}
    for v in S.preorder():
        if not v.parent:
            rates[v.ID] = baseline_rate
        else:
            var = lognormal_v * v.dist
            mu = np.log(rates[v.parent.ID]) - var/2                     # exp. value equal to parent's rate
            rates[v.ID] = np.exp(np.random.normal(mu, np.sqrt(var)))
            S_parent[v.ID] = v.parent.ID                                # save parent's ID for below
    
    for v in T.preorder():
        if v.parent:
            S_v_ID = v.color[1] if isinstance(v.color, tuple) else v.color
            effective_rate = (rates[S_parent[S_v_ID]] + rates[S_v_ID]) / 2
            v.dist = v.dist * effective_rate



# -*- coding: utf-8 -*-

"""
Tree Simulator.

Simulate species and gene trees.

Methods in this module:
    - build_species_tree
    - make_ultrametric
    - build_gene_tree
    - observable_tree
"""

import random
from collections import deque

import numpy as np

from asymmetree.tools.PhyloTree import PhyloTree, PhyloTreeNode


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2020, David Schaller"


# --------------------------------------------------------------------------
#                    CONSTRUCTION OF THE SPECIES TREE
#
#                       with the innovations model
# --------------------------------------------------------------------------

def build_species_tree(N, planted=True, model="innovations",
                       non_binary=0.0):
    """Builds a species tree S with N leaves.
    
    Keyword parameters:
        planted -- add a planted root that has the canonical root as its
                   single neighbor; default is True
        model -- simulation model to be applied; default is 'innovations'
        non_binary -- probality that an inner edge is contracted;
                      results in non-binary tree; default is 0.0
    """
    
    if isinstance(model, str) and model.lower() in ('innovation', 'innovations'):
        tree = _innovations_model(N, planted=planted)
    else:
        raise ValueError(f"Model '{model}' is not available!")
        
    if non_binary > 0.0:
         edges = _select_edges_for_contraction(tree,
                                               min(non_binary, 1.0),
                                               exclude_planted_edge=True)
         tree.contract(edges)
         
    make_ultrametric(tree)
        
    return tree
        

def _innovations_model(N, planted=True):
    """Builds a species tree S with N leaves with the innovations model."""
    
    tree = PhyloTree(PhyloTreeNode(0, label="0"))
    tree.number_of_species = N
    
    if planted:                     # planted tree (root is an implicit
                                    # outgroup with outdegree = 1)
        root = PhyloTreeNode(1, label="1")
        tree.root.add_child(root)
        node_counter = 2
    else:
        root = tree.root
        node_counter = 1
    
    t = 1
    features = [0]                  # set of available features
    species = {(0,): root}          # extant species
    
    while len(species) < N:
        
        loss_candidates = set()     # species for which loss of a feature
        for s in species.keys():    # can trigger a speciation
            for i in range(0, len(s)):
                    if s[:i] + s[i+1:] not in species:
                        loss_candidates.add(s)
        
        if not loss_candidates:     # INNOVATION EVENT
            s = random.choice(list(species))
            new_feature = len(features)
            
            new_s = s + (new_feature,)
            
            child1 = PhyloTreeNode(node_counter, label=str(node_counter))
            species[s].add_child(child1)
            child2 = PhyloTreeNode(node_counter+1, label=str(node_counter+1))
            species[s].add_child(child2)
            
            node_counter += 2
            
            species[s] = child1
            species[new_s] = child2
            features.append(new_feature)
            t += 1
            
        else:
            s = random.choice(list(loss_candidates))
            
            if len(s) > 1:
                feature_index = random.randint(0, len(s)-1)
            else:
                feature_index = 0
            
            new_s = s[:feature_index] + s[feature_index+1:]
            
            if new_s not in species:    # LOSS EVENT
                
                child1 = PhyloTreeNode(node_counter, label=str(node_counter))
                species[s].add_child(child1)
                child2 = PhyloTreeNode(node_counter+1, label=str(node_counter+1))
                species[s].add_child(child2)

                node_counter += 2
                
                species[s] = child1
                species[new_s] = child2
                t += 1
    
    return tree


def make_ultrametric(tree):
    """Makes a given species tree S ultrametric.
    
    It is t(root) = 1 and t(x) = 0 for x in L(S)."""
    for v in tree.preorder():
        if not v.parent:
            v.dist = 0.0
            v.tstamp = 1.0
        elif not v.children:
            v.dist = v.parent.tstamp
            v.tstamp = 0.0
        else:                               # random walk to a leaf
            pos = v                         # current position
            length = 0                      # path length |P|
            while pos.children:
                length += 1
                pos = pos.children[np.random.randint(len(pos.children))]
            v.dist = (v.parent.tstamp) * 2 * np.random.uniform() / (length+1)
            v.tstamp = v.parent.tstamp - v.dist
            
            
def _select_edges_for_contraction(tree, p, exclude_planted_edge=True):
    
    edges = []
    
    for u, v in tree.inner_edges():
        
        if exclude_planted_edge and (u is tree.root) and len(u.children) == 1:
            continue
        
        if random.random() < p:
            edges.append((u,v))
    
    return edges


# --------------------------------------------------------------------------
#                    CONSTRUCTION OF THE GENE TREE
#
# --------------------------------------------------------------------------
            
def build_gene_tree(S, DLH_rates):
    
    return _gillespie_direct(S, DLH_rates)

    
def _get_tstamp(t, total_rate):
    
    if total_rate <= 0.0:
        return -1
    else:
        return t - np.random.exponential(1/total_rate)


def _get_branch_and_type(DLH_rates, total_rate, branch_rates, i_to_b, ES_to_b):
    
    r = np.random.uniform() * total_rate
    index, current_sum = 0, 0
    
    for rate in branch_rates:
        if r <= current_sum + rate:
            break
        current_sum += rate
        index += 1
    
    branch = i_to_b[index]
    loss_factor = 1 if len(ES_to_b[(branch[2],branch[3])]) > 1 else 0
    
    if r <= current_sum + DLH_rates[0]:
        event_type = "D"
    elif r <= current_sum + DLH_rates[0] + loss_factor * DLH_rates[1]:
        event_type = "L"
    else:
        event_type = "H"
    
#    rate_sum = sum(branch_rates)
#    if total_rate != rate_sum:
#        raise KeyboardInterrupt
    
    return branch, event_type


def _coexisting_edges(sorted_edges, tstamp, exclude_edge=None):
    """Return list of edges for the given timestamp."""
    
    valid_edges = []
    for edge in sorted_edges:
        if edge[0].tstamp <= tstamp:
            break
        elif edge[1].tstamp < tstamp and edge != exclude_edge:
            valid_edges.append(edge)
    return valid_edges


def _gillespie_direct(S, DLH_rates):
    
    # --------------- initialization -----------------
    d, l, h = DLH_rates
    S_edges = S.sorted_edges()                      # edges of the species tree
                                                    # sort (u,v) by tstamp of u
    speciations = deque(S.sorted_nodes())           # queue for speciation events
    
    total_rate = 0                                  # total event rate (all branches)
    branch_rates = []                               # array for branch rates
    
    ES_to_b = {e: [] for e in S_edges}              # maps E(S) --> existing branches
    i_to_b = {}                                     # maps array index --> branch
    b_to_i = {}                                     # maps branch --> array index
    
    if len(S.root.children) > 1:
        T = PhyloTree(PhyloTreeNode(0, label="S",   # root is a speciation event
                                    color=S.root.ID, 
                                    dist=0.0, tstamp=1.0))
    else:                                           # planted tree
        T = PhyloTree(PhyloTreeNode(0, color=S.root.ID,
                                    dist=0.0, tstamp=1.0))
    speciations.popleft()
    
    id_counter = 1
    
    for S_v in S.root.children:
        new_branch = (id_counter, T.root, S.root, S_v, 0)
        ES_to_b[(S.root, S_v)].append(new_branch)
        index = len(branch_rates)
        branch_rates.append(d+h)
        total_rate += d+h
        i_to_b[index] = new_branch
        b_to_i[new_branch] = index
        id_counter += 1
        
    t = T.root.tstamp                               # start time = 1.0
    
    while speciations:
        
        event_tstamp = _get_tstamp(t, total_rate)
        
        # ----------------- SPECIATION -----------------
        if event_tstamp <= speciations[0].tstamp:
            S_v = speciations.popleft()
            S_u = S_v.parent
            
            for branch in ES_to_b[(S_u, S_v)]:
                b_id, b_parent, _, _, b_transferred = branch
                spec_node = PhyloTreeNode(b_id, label="S",
                                     color=S_v.ID, tstamp=S_v.tstamp,
                                     dist=abs(S_v.tstamp-b_parent.tstamp),
                                     transferred=b_transferred)
                b_parent.add_child(spec_node)
                if not S_v.children:
                    spec_node.label = str(spec_node.ID)
                for S_w in S_v.children:
                    new_branch = (id_counter, spec_node, S_v, S_w, 0)
                    ES_to_b[(S_v, S_w)].append(new_branch)
                    if S_w is S_v.children[0]:
                        index = b_to_i[branch]
                    else:
                        index = len(branch_rates)
                        branch_rates.append(branch_rates[b_to_i[branch]])
                        total_rate += branch_rates[-1]
                    i_to_b[index] = new_branch
                    b_to_i[new_branch] = index
                    id_counter += 1
                del b_to_i[branch]
                    
                
            t = S_v.tstamp
            
        # ----------------- D / L / H ------------------
        else:
            branch, event_type = _get_branch_and_type(DLH_rates, total_rate,
                                                      branch_rates, i_to_b, ES_to_b)
            b_id, b_parent, S_u, S_v, b_transferred = branch
            
            # ----------------- DUPLICATION -----------------
            if event_type == "D":
                dupl_node = PhyloTreeNode(b_id, label="D",
                                          color=(S_u.ID,S_v.ID),
                                          tstamp=event_tstamp,
                                          dist=abs(event_tstamp-b_parent.tstamp),
                                          transferred=b_transferred)
                b_parent.add_child(dupl_node)
                ES_to_b[(S_u, S_v)].remove(branch)
                for i in range(2):
                    new_branch = (id_counter, dupl_node, S_u, S_v, 0)
                    ES_to_b[(S_u, S_v)].append(new_branch)
                    if i == 0:
                        index = b_to_i[branch]
                        branch_rates[index] = d + l + h
                    else:
                        index = len(branch_rates)
                        branch_rates.append(d+l+h)
                    i_to_b[index] = new_branch
                    b_to_i[new_branch] = index
                    id_counter += 1
                del b_to_i[branch]
                
                if len(ES_to_b[(S_u, S_v)]) == 2:
                    total_rate += (d + l + h + l)
                else:
                    total_rate += (d + l + h)
            
            # -------------------- LOSS ---------------------
            elif event_type == "L":
                loss_node = PhyloTreeNode(b_id, label="*",
                                          color=(S_u.ID,S_v.ID),
                                          tstamp=event_tstamp,
                                          dist=abs(event_tstamp-b_parent.tstamp),
                                          transferred=b_transferred)
                b_parent.add_child(loss_node)
                ES_to_b[(S_u, S_v)].remove(branch)
                branch_rates[b_to_i[branch]] = 0
                del b_to_i[branch]
                
                if len(ES_to_b[(S_u, S_v)]) == 1:
                    total_rate -= (d + l + h + l)
                    branch_rates[b_to_i[ES_to_b[(S_u, S_v)][0]]] -= l
                else:
                    total_rate -= (d + l + h)
            
            # -------------------- HGT ----------------------
            elif event_type == "H":
                valid_edges = _coexisting_edges(S_edges, event_tstamp,
                                                exclude_edge=(S_u, S_v))
                if valid_edges:
                    trans_edge = random.choice(valid_edges)
                    hgt_node = PhyloTreeNode(b_id, label="H",
                                             color=(S_u.ID,S_v.ID),
                                             tstamp=event_tstamp,
                                             dist=abs(event_tstamp-b_parent.tstamp),
                                             transferred=b_transferred)
                    b_parent.add_child(hgt_node)
                    ES_to_b[(S_u, S_v)].remove(branch)
                    
                    new_branch = (id_counter, hgt_node, S_u, S_v, 0)        # original branch
                    ES_to_b[(S_u, S_v)].append(new_branch)
                    index = b_to_i[branch]
                    i_to_b[index] = new_branch
                    b_to_i[new_branch] = index
                    id_counter += 1
                    
                    trans_branch = (id_counter, hgt_node, *trans_edge, 1)   # receiving branch
                    ES_to_b[trans_edge].append(trans_branch)
                    index = len(branch_rates)
                    branch_rates.append(d+l+h)
                    i_to_b[index] = trans_branch
                    b_to_i[trans_branch] = index
                    id_counter += 1
                    
                    del b_to_i[branch]
                    
                    if len(ES_to_b[trans_edge]) == 2:
                        total_rate += (d + l + h + l)
                        branch_rates[b_to_i[ES_to_b[trans_edge][0]]] += l
                    else:
                        total_rate += (d + l + h)
            
            t = event_tstamp

#    # check that there is no extinction in any species
#    VS_to_VT = {l.ID: [] for l in S.preorder() if not l.children}
#    for v in T.preorder():
#        if not v.children and not v.label == "*":
#            VS_to_VT[v.color].append(v.ID)
#    for leaf_list in VS_to_VT.values():
#        if not leaf_list:
#            raise KeyboardInterrupt
    
    return T


# --------------------------------------------------------------------------
#                 CONSTRUCTION OF THE OBSERVABLE TREE
#
# --------------------------------------------------------------------------
    
def observable_tree(tree):
    obs_tree = tree.copy()
    
    loss_nodes = []
    for node in obs_tree.postorder():
        if not node.children and node.label == "*":
            loss_nodes.append(node)
    
    # traverse from loss node to root delete if degree <= 1
    for loss_node in loss_nodes:
        current = obs_tree.delete_and_reconnect(loss_node,
                                                add_distances=True,
                                                keep_transferred=True)
        
        while len(current.children) < 2 and current.parent:
            current = obs_tree.delete_and_reconnect(current,
                                                    add_distances=True,
                                                    keep_transferred=True)
    
    # delete the root if the tree is planted
    if len(obs_tree.root.children) == 1:
        obs_tree.delete_and_reconnect(obs_tree.root.children[0],
                                      add_distances=False,
                                      keep_transferred=False)
    
    return obs_tree


# -*- coding: utf-8 -*-

"""
File I/O.
"""

import itertools

import networkx as nx

__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


def matrix_to_phylip(filename, leaves, matrix):
    """Write the distance matrix in phylip format."""
    with open(filename, "w") as f:
        f.write(str(len(leaves)))
        for i in range(len(leaves)):
            f.write("\n" + str(leaves[i].ID))
            for j in range(len(leaves)):
                f.write("\t" + str(matrix[i,j]))


def species_to_genes(filename, scenario):
    """Write the species and corresponding genes file."""
    with open(filename, "w") as f:
        species = list(scenario.color_dict)
        for i in range(len(species)):
            f.write(str(species[i]))
            for gene in scenario.color_dict[species[i]]:
                f.write("\t" + str(gene.ID))
            if i < len(species)-1:
                f.write("\n")


def species_pairs_outgroups(filename, scenario):
    """Write pairs of species with corresponding outgroups 
    
    ("ok-species" file)."""
    with open(filename, "w") as f:
        begin = True
        for i in range(len(scenario.subtree_list)):
            for speciesX, speciesY in itertools.combinations(scenario.subtree_list[i], 2):
                if begin:
                    f.write(str(speciesX) + "\t" + str(speciesY))
                    begin = False
                else:
                    f.write("\n" + str(speciesX) + "\t" + str(speciesY))
                for j in range(len(scenario.subtree_list)):
                    if i == j:
                        continue
                    for outgroup_species in scenario.subtree_list[j]:
                        f.write("\t" + str(outgroup_species))
   
                 
def species_in_subtree(filename, scenario):
    """Species in the same subtree under the the root
    
    (one line per subtree)."""
    with open(filename, "w") as f:
        begin = True
        for sp_list in scenario.subtree_list:
            if begin:
                f.write("\t".join([str(item) for item in sp_list]))
                begin = False
            else:
                f.write("\n")
                f.write("\t".join([str(item) for item in sp_list]))


def subtree_BMG(filename, scenario):
    """Write the (subgraph of) BMG."""
    with open(filename, "w") as f:
        begin = True
        for u, v in scenario.BMG_subtrees.edges:
            if begin:
                f.write(str(u) + "\t" + str(v))
                begin = False
            else:
                f.write("\n" + str(u) + "\t" + str(v))


def parse_BMG_edges(output, scenario):
    """Parse edge output from external program to a graph
    
    (converts keys to ints if possible)."""
    BMG = nx.DiGraph()
    BMG.add_nodes_from(scenario.BMG.nodes(data=True))
    
    for line in output.split("\n"):
        if line:
            edge = line.split()
            if len(edge) >= 2:
                u = int(edge[0]) if edge[0].isdigit() else edge[0]
                v = int(edge[1]) if edge[1].isdigit() else edge[1]
                if v not in BMG.nodes:
                    print("Not in BMG", v)
                if u not in BMG.nodes:
                    print("Not in BMG", u)
                BMG.add_edge(u, v)
    return BMG


def write_newick(filename, tree):
    """Write Newick tree."""
    with open(filename, "w") as f:
        f.write(tree.to_newick())


if __name__ == "__main__":
    
    # generate new files in folder "test_data"
    
    import asymmetree.simulator.TreeSimulator as ts
    import asymmetree.simulator.TreeImbalancer as tm
    
    from asymmetree.simulator.Scenario import Scenario
    from asymmetree.tools.PhyloTree import PhyloTree
    
    matrix_file = "test_data/matrix.phylip"
    species_file = "test_data/spec_genes.txt"
    outgroup_file = "test_data/outgroups.txt"
    subtrees_file = "test_data/species_subtrees.txt"
    bmg_file = "test_data/bmg.txt"
    tree_file = "test_data/species_tree.txt"
    
    DLH_rates = (1,1,0)
    
    #S = ts.build_species_tree(10, planted=True)
    S = PhyloTree.parse_newick("(((8:0.08603999468839801,(10:0.06055381385164242,(12:0.02750356935270675,(14:0.0071494825768602215,15:0.0071494825768602215)13:0.02035408677584653)11:0.03305024449893567)9:0.025486180836755596)2:0.34305906001624026,(((16:0.036223587639635554,(18:0.032127988304223636,19:0.032127988304223636)17:0.0040955993354119214)6:0.1114990457717915,7:0.14772263341142705)4:0.03331012904283408,5:0.18103276245426114)3:0.24806629225037713)1:0.5709009452953617)0:0.0")
    S.reconstruct_IDs()
    S.reconstruct_timestamps()
    print("------------- S -------------")
    print(S.to_newick())
    
    TGT = ts.build_gene_tree(S, DLH_rates)
    print("done")
    TGT = tm.imbalance_tree(TGT, S, baseline_rate=1,
                                  lognormal_v=0.2,
                                  gamma_param=(0.5, 1.0, 2.2),
                                  weights=(1/3, 1/3, 1/3),
                                  copy_tree=False)
    
    print("------------- TGT -------------")
    print(TGT.to_newick(distance_only=False))
    
    scenario = Scenario(S, TGT, DLH_rates)
    print(scenario.subtree_list)
    D = scenario.get_distance_matrix()
    
    # write the distance matrix
    matrix_to_phylip(matrix_file, scenario.genes, D)
    # write the species and corresponding genes file
    species_to_genes(species_file, scenario)
    # write the "ok-species" files
    species_pairs_outgroups(outgroup_file, scenario)      
    # write species in root subtrees
    species_in_subtree(subtrees_file, scenario)
    # write the BMG
    subtree_BMG(bmg_file, scenario)
    # write species tree to newick format
    write_newick(tree_file, S)
    
    # test reading input
    with open(bmg_file, "r") as f:
        BMG = parse_BMG_edges(f.read(), scenario)
        #print(BMG.edges)


# -*- coding: utf-8 -*-

""" 
Parser for ProteinOrtho files.

Reads two different versions of ProteinOrtho result files (with/without
'-synteny' option) and builds a graph from the data (color --> 
species/filename).

References:
    - Marcus Lechner, Sven Findeiß, Lydia Steiner, Manja Marz, Peter F.
      Stadler, Sonja J. Prohaska. Proteinortho: Detection of (Co-)orthologs
      in large-scale analysis. BMC Bioinformatics 2011, 12:124
"""

import time
from multiprocessing import Pool

import networkx as nx

from asymmetree.proteinortho.SpeciesTreeFromParalogs import TreeReconstructor   


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


def parse_po_graph(filename):
    
    G = nx.Graph()
    
    with open(filename, "r") as f:
        
        # skip the first two lines
        f.readline(); f.readline
        
        col_a, col_b= "unknown", "unknown"
        line = f.readline().strip()
        
        while line:
            
            line = line.strip()
            
            # new color pair begins
            if line and line.startswith("#") and not line.startswith("# Scores:"):
                colors = line.split()
                if len(colors) == 3:
                    col_a, col_b = colors[1], colors[2]
                
            elif line and not line.startswith("#"):
                edge = line.split()
                id1 = "{}_{}".format(col_a, edge[0])
                id2 = "{}_{}".format(col_b, edge[1])
                if id1 not in G:
                    G.add_node(id1, color=col_a)
                if id2 not in G:
                    G.add_node(id2, color=col_b)
                    
                # file type 1 with simscore (synteny was activated)
                if len(edge) == 8:
                    G.add_edge(id1, id2,
                               evalue_ab = float(edge[2]), 
                               bitscore_ab = float(edge[3]),
                               evalue_ba = float(edge[4]),
                               bitscore_ba = float(edge[5]),
                               same_strand = int(edge[6]),
                               simscore = float(edge[7]))
                
                # file type 2 without simscore
                elif len(edge) == 6:
                    G.add_edge(id1, id2,
                               evalue_ab = float(edge[2]), 
                               bitscore_ab = float(edge[3]),
                               evalue_ba = float(edge[4]),
                               bitscore_ba = float(edge[5]))
                else:
                    print("Check file format!", edge)
                    continue
                
            line = f.readline()
            
    return G


def parse_best_match_candidates(filename):
    
    candidates = {}
    
#    # extract the species from filename
#    species_match = re.search(r"([^/]+)\.bm_candidates$", filename)
#    if species_match:
#        species = species_match.group(1)
#    else:
#        raise ValueError(f"Could not extract species from filename '{filename}'.")
    
    with open(filename, "r") as f:
        
        line = f.readline()
        
        while line:
            
            line = line.strip()
            if not line:
                line = f.readline()
                continue
            
            data = line.split()
            
            if len(data) != 6:
                print(f"Check file format: {filename}, line {line}")
                continue
                
            color      = data[0]
            query_id   = data[1]
            subject_id = data[2]
            e_value    = float(data[3])
            bitscore   = float(data[4])
            identity   = float(data[5]) / 100
            
            if subject_id not in candidates:
                candidates[subject_id] = {}
            if color not in candidates[subject_id]:
                candidates[subject_id][color] = []
            candidates[subject_id][color].append( (color, query_id, e_value, bitscore, identity) )
                
            line = f.readline()
            
    return candidates


def parse_best_matches(filename):
    
    G = nx.DiGraph()
    
    with open(filename, "r") as f:
        
        line = f.readline()
        
        while line:
            
            line = line.strip().split()
            
            if len(line) == 4:
                
                col1, col2 = line[2], line[3]
                id1 = "{}_{}".format(col1, line[0])
                id2 = "{}_{}".format(col2, line[1])
                if id1 not in G:
                    G.add_node(id1, color=col1)
                if id2 not in G:
                    G.add_node(id2, color=col2)
                G.add_edge(id1, id2)
            
            line = f.readline()
            
    return G


def reconstruct(params):
    
    G, cotree_mode, triple_mode = params
    
    start_time = time.time()
    tr = TreeReconstructor(cotree_mode=cotree_mode)
    tr.add_ortho_graph(G)
    tr.build_species_tree(mode=triple_mode)
    newick = tr.newick_with_support()
    
    return newick, time.time() - start_time


def reconstruct_trees_and_write(infile, outfile, cotree_modes, triple_modes,
                                parallel_processing=True):
    
    G = parse_po_graph(infile)
    
    inputs = [(G, m1, m2) for m1 in cotree_modes for m2 in triple_modes]
    
    if parallel_processing:
        with Pool() as p:
            results = p.map(reconstruct, inputs)
    else:
        results = list(map(reconstruct, inputs))
    
    with open(outfile, "w") as f:
        for i in range(len(inputs)):
            
            _, cotree_mode, triple_mode = inputs[i]
            newick, time_needed = results[i]
            f.write(f"# Cotree usage mode: {cotree_mode}, "\
                    f"Max. Consistent Triple Set heuristic: {triple_mode}, "\
                    f"Time: {time_needed}\n")
            f.write(newick + "\n")


if __name__ == "__main__":
    
    import os
    
    directory = "../../testing/test_files_2"
    
    infile = os.path.join(directory, "test.proteinortho-graph")
    outfile = os.path.join(directory, "species_trees_heuristic")
    
    cotree_modes = ["best", "all"]
    triple_modes = ["Mincut", "BPMF", "Greedy"]
    
    reconstruct_trees_and_write(infile, outfile,
                                cotree_modes, triple_modes,
                                parallel_processing=True)

    with open(outfile, "r") as f:
        print(f.read())
    
#    tr = TreeReconstructor(cotree_mode="all")
#    tr.add_ortho_graph(G)
#    spec_tree = tr.build_species_tree(mode="BPMF")
#    
#    print(spec_tree.to_newick())
#    print(tr.newick_with_support())
#    
#    G = parse_best_matches("test_files_2/bmg")
#    print(G.order(), G.size())


# -*- coding: utf-8 -*-

import itertools

import numpy as np

from asymmetree.tools.PhyloTree import PhyloTree, PhyloTreeNode


class OutgroupFinder:
    
    def __init__(self, tree_file):
        
        self.tree_file = tree_file
        
        self._prepare()
        
        
    def _prepare(self):
        
        self._parse_tree()
        
        self._compute_lcas_in_S()
        self._compute_outgroup_species()
      
        
    def _parse_tree(self):
        
        with open(self.tree_file, "r") as f:
            newick = f.readline()
        
        self.S = PhyloTree.parse_newick(newick)
        
        
    def _compute_lcas_in_S(self):
        
        self.S.supply_leaves()
        N = len(self.S.root.leaves)
        self.leaf_index = {leaf.label: i for i, leaf in enumerate(self.S.root.leaves)}
        
        self.lca = np.zeros((N,N), dtype=PhyloTreeNode)
        self.subtree_species = {}
        
        for v in self.S.preorder():
            self.subtree_species[v] = [leaf.label for leaf in v.leaves]
            
            if not v.children:
                self.lca[self.leaf_index[v.label],self.leaf_index[v.label]] = v
            elif len(v.children) >= 2:
                for c1, c2 in itertools.combinations(v.children, 2):
                    for x in c1.leaves:
                        x_index = self.leaf_index[x.label]
                        for y in c2.leaves:
                            y_index = self.leaf_index[y.label]
                            self.lca[x_index, y_index] = v
                            self.lca[y_index, x_index] = v
    
    
    def _compute_outgroup_species(self):
        
        self.lca_to_outgroups = {key: set() for key in self.subtree_species.keys()}
        
        for v in self.S.preorder():
            if v.parent:
                self.lca_to_outgroups[v].update(self.lca_to_outgroups[v.parent])
                
            for c1, c2 in itertools.permutations(v.children, 2):
                self.lca_to_outgroups[c1].update(self.subtree_species[c2])
                
    
    def __call__(self, spec_x, spec_Y, candidates, limit=10):
        
        # set of candidates might be empty (?)
        if not candidates:
            return False
        
        outgroup_spec = self.lca_to_outgroups[self.lca[self.leaf_index[spec_x],
                                                       self.leaf_index[spec_Y]]]
        
        # set of potential outgroup species can be empty (when lca is the root)
        if not outgroup_spec:
            return False
        
        available_outgroup_spec = outgroup_spec.intersection(candidates)
        
        total_outgroup_nr = 0
        for spec in available_outgroup_spec:
            total_outgroup_nr += len(candidates[spec])
            
        # set of actual outgroup genes can be empty
        if total_outgroup_nr == 0:
            return False
        
        outgroups = []
        
        # pick one outgroup from each species in each round
        i, outgroup_counter, stop = 0, 0, min(limit, total_outgroup_nr)
        while outgroup_counter < stop:
            
            for spec in available_outgroup_spec:
                if i < len(candidates[spec]):
                    outgroups.append( candidates[spec][i] )
                    outgroup_counter += 1
                    if outgroup_counter == stop:
                        break
            i += 1
        
        return outgroups


# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


class GeneTreeVis:
    
    def __init__(self, tree):
        
        self.tree = tree
        
        self.symbolsize = 0.03
        self.symbollw = 0.04
        self.leafs_per_vertical_unit = 15
        self.symbol_zorder = 3        
        
        print(tree.to_newick())
        
        self.distance_dict = {}
        self.colors = {}
        #self.divtime_dict {}
        self.leaf_counter = 0
        self.node_positions = {}
        
        self.draw()
        
    
    def draw(self):
        
        self.fig, self.ax = plt.subplots()
        self.ax.set_aspect('equal')
        self.ax.invert_yaxis()
        
        self.initial_traversal()
        self.assign_colors()
        self.assign_positions()
        self.draw_edges()
        self.draw_nodes()
        
        self.ax.axvline(x=0, linewidth=1, color='grey', linestyle='--')
        self.ax.set_yticks([])
        self.ax.spines["top"].set_visible(False)
        self.ax.spines["left"].set_visible(False)
        self.ax.spines["right"].set_visible(False)
        
        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()
        self.fig.set_size_inches(5*abs(xmax-xmin), 5*abs(ymax-ymin)+0.4)
        plt.tight_layout()
        plt.show()
        
    
    def initial_traversal(self, distance="evolutionary"):
        
        xmax = 0.0
        
        for v in self.tree.preorder():
            if distance == "evolutionary":
                if not v.parent:
                    self.distance_dict[v] = 0.0
                else:
                    self.distance_dict[v] = self.distance_dict[v.parent] + v.dist
                    if self.distance_dict[v] > xmax:
                        xmax = self.distance_dict[v]
            elif distance == "divtime":
                raise ValueError("Divergence time not yet implemented!")
            else:
                raise ValueError("Distance mode not known: " + str(distance))
            if not v.children:
                self.leaf_counter += 1
                if v.label != '*' and v.color not in self.colors:
                    self.colors[v.color] = len(self.colors)
                
        self.ax.set_xlim(-0.1,xmax+0.5)
        
    
    def assign_colors(self):
        
        if len(self.colors) <= 10:
            cmap = plt.get_cmap('tab10')(np.arange(len(self.colors), dtype=int))
        else:
            cmap = plt.get_cmap('jet')(np.linspace(0, 1.0, len(self.colors)))
        for color_label, color in zip(self.colors.keys(), cmap):
            self.colors[color_label] = color
        print(self.colors)
    
    
    def assign_positions(self):
        
        ymax = (self.leaf_counter-1)/self.leafs_per_vertical_unit
        self.ax.set_ylim(ymax+0.1, -self.symbolsize*0.6)
        
        yposition = 0
        for v in self.tree.postorder():
            if not v.children:
                self.node_positions[v] = (self.distance_dict[v],
                                          yposition)
                yposition += 1/self.leafs_per_vertical_unit
            else:
                ymean = (self.node_positions[v.children[0]][1] +
                         self.node_positions[v.children[-1]][1])/2
                self.node_positions[v] = (self.distance_dict[v],
                                          ymean)
    
    
    def draw_edges(self):
        
        for v in self.tree.preorder():
            if v.parent:
                self.ax.plot([self.node_positions[v.parent][0],
                              self.node_positions[v][0]],
                             [self.node_positions[v][1],
                              self.node_positions[v][1]],
                     color='black',
                     linestyle='-', linewidth=1)
            if v.children:
                self.ax.plot([self.node_positions[v][0],
                              self.node_positions[v][0]],
                             [self.node_positions[v.children[0]][1],
                              self.node_positions[v.children[-1]][1]],
                     color='black',
                     linestyle='-', linewidth=1)
    
    
    def draw_nodes(self):
        
        for v in self.tree.preorder():
            if not v.parent:
                self.draw_root(*self.node_positions[v])
            elif v.children:
                if v.label == 'D':
                    self.draw_dupl(*self.node_positions[v])
                elif v.label == 'S':
                    self.draw_spec(*self.node_positions[v])
                if v.label == 'H':
                    self.draw_hgt(*self.node_positions[v])
            else:
                if v.label == '*':
                    self.draw_loss(*self.node_positions[v])
                else:
                    x, y = self.node_positions[v]
                    self.draw_leaf(x, y,
                                   color=self.colors[v.color])
                    self.write_label(x+self.symbolsize+0.02, y,
                                     str(v.label))
                
    
    def draw_leaf(self, x, y, color='white', leftalign=True):
        
        if leftalign:
            x += self.symbolsize/2
        fill = mpatches.Circle((x, y), self.symbolsize/2,
                               color=color, fill=True,
                               zorder=self.symbol_zorder)
        self.ax.add_patch(fill)
        outer = mpatches.Circle((x, y), self.symbolsize/2,
                                color='black', fill=False,
                                lw=self.symbolsize/self.symbollw,
                                zorder=self.symbol_zorder)
        self.ax.add_patch(outer)
        little = mpatches.Circle((x, y), self.symbolsize/8,
                                 color='black', fill=True,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(little)
        
    
    def draw_loss(self, x, y):
        
        self.ax.plot([x,x],
                     [y-self.symbolsize/2, y+self.symbolsize/2],
                     color='black',
                     linestyle='-', linewidth=1)
    
    
    def draw_root(self, x, y, rightalign=True):
        
        if rightalign:
            x -= self.symbolsize/2
        fill = mpatches.Circle((x, y), self.symbolsize/2,
                               color='white', fill=True,
                               zorder=self.symbol_zorder)
        self.ax.add_patch(fill)
        outer = mpatches.Circle((x, y), self.symbolsize/2,
                                color='black', fill=False,
                                lw=self.symbolsize/self.symbollw,
                                zorder=self.symbol_zorder)
        self.ax.add_patch(outer)
        little = mpatches.Circle((x, y), self.symbolsize/5,
                                 color='black', fill=False,
                                 lw=self.symbolsize/self.symbollw,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(little)
    
    
    def draw_spec(self, x, y):
        
        fill = mpatches.Circle((x, y), self.symbolsize/2,
                               color='black', fill=True,
                               zorder=self.symbol_zorder)
        self.ax.add_patch(fill)
        
    
    def draw_dupl(self, x, y):
        
        square = mpatches.Rectangle((x-self.symbolsize/2,
                                     y-self.symbolsize/2),
                                    width=self.symbolsize,
                                    height=self.symbolsize,
                                    color='white', fill=True,
                                    zorder=self.symbol_zorder)
        self.ax.add_patch(square)
        border = mpatches.Rectangle((x-self.symbolsize/2,
                                     y-self.symbolsize/2),
                                    width=self.symbolsize,
                                    height=self.symbolsize,
                                    color='black', fill=False,
                                    lw=self.symbolsize/self.symbollw,
                                    zorder=self.symbol_zorder)
        self.ax.add_patch(border)
    
    
    def draw_hgt(self, x, y):
        
        coord = np.asarray([[x,y-self.symbolsize/2],
                            [x+self.symbolsize/2,y+self.symbolsize/2],
                            [x-self.symbolsize/2,y+self.symbolsize/2]])

        inner = mpatches.Polygon(coord, closed=True,
                                 color='white', fill=True,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(inner)
        outer = mpatches.Polygon(coord, closed=True,
                                 color='black', fill=False,
                                 lw=self.symbolsize/self.symbollw,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(outer)
    
    
    def write_label(self, x, y, text):
        
        self.ax.text(x, y, text,
                     horizontalalignment='left',
                     verticalalignment='center')


if __name__ == "__main__":
    
    import asymmetree.simulator.TreeSimulator as ts
    import asymmetree.simulator.TreeImbalancer as tm
    
    S = ts.build_species_tree(6, planted=True, non_binary=0.2)
    
    TGT = ts.build_gene_tree(S, (1.0,1.0,1.0))
    TGT = tm.imbalance_tree(TGT, S, baseline_rate=1,
                            lognormal_v=0.2,
                            gamma_param=(0.5, 1.0, 2.2),
                            weights=(1/3, 1/3, 1/3),
                            copy_tree=False)
    
    gtv = GeneTreeVis(TGT)


# -*- coding: utf-8 -*-

"""
Distance Matrix.

Computation of the distance matrix from a gene tree.

Methods in this module:
    - distance_matrix
"""

import itertools
import numpy as np


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


def distance_matrix(tree, leaves=None, leaf_index=None):
    """Compute the distance matrix from a gene tree.
    
    Additionally a list of leaves and a dictionary containing
    the indices as values is computed (if not supplied)
    and returned.
    
    Keyword arguments:
        leaves -- set of leaves.
        leaf_index -- dictionary having the leaves as keys
                      and matrix indices as values.
    """
    distance_dict, _ = tree.distances_from_root()
    
    if not leaves:
        tree.supply_leaves()
        color_dict = {}
        for leaf in tree.root.leaves:
            if leaf.color not in color_dict:
                color_dict[leaf.color] = []
            color_dict[leaf.color].append(leaf)
        leaves = []                                                 # color-sorted leaf list
        for color, leaf_list in color_dict.items():
            for leaf in leaf_list:
                leaves.append(leaf)
    
    if not leaf_index:                                              # maps leaf to index in matrix
        leaf_index = {leaf: i for i, leaf in enumerate(leaves)}
    
    D = np.zeros((len(leaves),len(leaves)), dtype=np.float)
    
    for v in tree.preorder():
        if v.children:
            for c1, c2 in itertools.combinations(v.children, 2):
                for x in c1.leaves:
                    x_index = leaf_index[x]
                    x_dist = distance_dict[x] - distance_dict[v]
                    for y in c2.leaves:
                        y_index = leaf_index[y]
                        y_dist = distance_dict[y] - distance_dict[v]
                        D[x_index, y_index] = x_dist + y_dist
                        D[y_index, x_index] = x_dist + y_dist
    
    return leaves, leaf_index, D


# -*- coding: utf-8 -*-

"""
Extended Best Hits Method.

Implementation of the Extended Best Hits method for best match inference.

Methods in this module:
    - ebh_qinfer
    - ebh_from_scenario
    - ebh
"""

import os, subprocess, time
import networkx as nx

from asymmetree.tools import FileIO
from asymmetree.best_matches import TrueBMG


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"



# --------------------------------------------------------------------------
#                           PYTHON IMPLEMENTATION
#
# --------------------------------------------------------------------------
    
def ebh(leaves, D, epsilon=0.000_000_01):
    """Compute BMG and RBMG from a distances matrix D.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM threshold: (x,y) in BMG if
                   D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=10E-8 (for limited float precision).
    """
    BMG, RBMG = nx.DiGraph(), nx.Graph()
    colors = set()
    relative_threshold = 1 + epsilon
    
    for v in leaves:
        BMG.add_node(v.ID, label=v.label, color=v.color)
        RBMG.add_node(v.ID, label=v.label, color=v.color)
        colors.add(v.color)
    
    # ---- build BMG ----
    for u in range(len(leaves)):
        minima = {color: float('inf') for color in colors}
        for v in range(len(leaves)):
            if D[u,v] < minima[leaves[v].color]:
                minima[leaves[v].color] = D[u,v]
        for v in range(len(leaves)):
            if (leaves[u].color != leaves[v].color and
                D[u,v] <= relative_threshold * minima[leaves[v].color]):
                BMG.add_edge(leaves[u].ID, leaves[v].ID, distance = D[u,v])
    
    # ---- build RBMG as symmetric part of the BMG ----
    for x, neighbors in BMG.adjacency():
        for y in neighbors:
            if BMG.has_edge(y,x):
                RBMG.add_edge(x,y)
    
    return BMG, RBMG


# --------------------------------------------------------------------------
#                            EXTERNAL C++ PROGRAM
#
# --------------------------------------------------------------------------
 
def ebh_qinfer(scenario,
               matrix_filename, species_filename,
               epsilon=0.5,
               benchmark_file=None,
               binary_path=None):
    """Compute BMG and RBMG from a distances matrix D using 'qinfer'.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM threshold: (x,y) in BMG if
                   D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=10E-8 (for limited float precision).
        benchmark_file -- activate benchmarking in 'qinfer' and
                          specify the filename
        binary_path -- path to 'qinfer' binary (if not available
                       within path)
    """
    
    if not binary_path:
        qinfer_command = "qinfer"
    elif os.path.exists(binary_path):
        qinfer_command = binary_path
    else:
        raise FileNotFoundError(f"Path to qinfer binary file '{binary_path}' does not exist!")
    
    output = -1
    command = [qinfer_command, matrix_filename, species_filename,
               "--disable-quartet", "--epsilon=" + str(epsilon)]

    if benchmark_file is not None:
        command.append( "--benchmark=" + benchmark_file )
    
    # call 'qinfer' and measure execution time
    start = time.time()
    
    try:
        output = subprocess.run(command, stdout=subprocess.PIPE)
    except:
        raise FileNotFoundError("Calling qinfer failed!")
    
    exec_time = time.time() - start
    
    if output == -1:
        raise Exception("No output from qinfer!")
    
    BMG = FileIO.parse_BMG_edges(output.stdout.decode(), scenario)
    RBMG = TrueBMG.RBMG_from_BMG(BMG)
    
    return BMG, RBMG, exec_time


def ebh_from_scenario(scenario, epsilon=0.5):
    """Compute BMG and RBMG from a scenario using 'qinfer'.
    
    Keyword arguments:
        epsilon -- epsilon for relative BM threshold: (x,y) in BMG if
                   D(x,y) <= (1+epsilon) * min d(x,y'),
                   default=10E-8 (for limited float precision).
    """

    matrix_filename = "temp.phylip"
    species_filename = "temp_species.txt"
    
    matrix = scenario.get_distance_matrix()
    FileIO.matrix_to_phylip(matrix_filename, scenario.genes, matrix)
    FileIO.species_to_genes(species_filename, scenario)
    
    BMG, RBMG, exec_time = ebh_qinfer(scenario,
                                      matrix_filename, species_filename,
                                      epsilon=epsilon)
    
    os.remove(matrix_filename)
    os.remove(species_filename)
    
    return BMG, RBMG, exec_time